#include<sys/types.h>
#include<sys/stat.h>
#include<sys/prctl.h>
#include<openssl/evp.h>
#include<bsd/stdlib.h>
#include<bsd/unistd.h>
#include<openssl/blowfish.h>

#include<errno.h>
#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<fcntl.h>
#include<unistd.h>


/**
*This is a function that alerts the user to how
*to use the function.  If this function is called,
*a failure condition has been met.
**/
void help()
{
	printf("Usage: bfed [-deh] -k key\n");
	exit(EXIT_FAILURE);
}


/**
*This is the encryption/decryption function.  It is adapted
*from the man page for EVP_Encryptinit.  I added error checking
*to the function as well, declaration of variables, and a vraible
*key input by the user.
**/
int do_crypt(FILE *in, FILE *out, int do_encrypt, unsigned char* key)
{
        /* Allow enough space in output buffer for additional block */
        unsigned char inbuf[1024], outbuf[1024 + EVP_MAX_BLOCK_LENGTH];
	int inlen, outlen;
	EVP_CIPHER_CTX ctx;
       	/* Bogus key and IV: we'd normally set these from
         * another source.
         */
        unsigned char iv[] = "00000000";
        /* Don't set key or IV because we will modify the parameters */
        EVP_CIPHER_CTX_init(&ctx);

	/*Initializes the current structure.*/
        if(EVP_CipherInit_ex(&ctx, EVP_bf_cbc(), NULL, NULL, NULL, do_encrypt) == 0)
	{
		fprintf(stderr, "%s: Error initializing the structure.\n", getprogname());
		EVP_CIPHER_CTX_cleanup(&ctx);
		memset(key,0,16);
		exit(EXIT_FAILURE);
	}

	/*Sets the key length that the user entered.*/
        if(EVP_CIPHER_CTX_set_key_length(&ctx, 16) == 0)
        {
                fprintf(stderr, "%s: Error setting the key length\n.", getprogname());
                EVP_CIPHER_CTX_cleanup(&ctx);
                memset(key,0,16);
                exit(EXIT_FAILURE);
        }

        /* We finished modifying parameters so now we can set key and IV */
        if(EVP_CipherInit_ex(&ctx, NULL, NULL, key, iv, do_encrypt) == 0)
        {
                fprintf(stderr, "%s: Error re-initializing the structure.\n", getprogname());
                EVP_CIPHER_CTX_cleanup(&ctx);
                memset(key,0,16);
                exit(EXIT_FAILURE);
        }

	/*Inited loop to allow for reading.*/
        for(;;)
        {
		/*Read info in until end of file or ^D^D^D are endtered. */
        	inlen = fread(inbuf, 1, 1024, in);
		/*Either an error occurred or EOF was reached so break the loop.*/
                if(inlen <= 0) break;
		/*Get Ciphertext for what was read.*/
                if(!EVP_CipherUpdate(&ctx, outbuf, &outlen, inbuf, inlen))
                {
			fprintf(stderr, "%s: Error updating the structure.\n", getprogname());
                	/* Cleans up the structure, erases key*/
                        EVP_CIPHER_CTX_cleanup(&ctx);
			memset(key, 0, 16);
                        exit(EXIT_FAILURE);
                }
                
		/*Writes ciphertext to File*.*/
		fwrite(outbuf, 1, outlen, out);
        }
        
	/*Finalizes the structure.*/
	if(!EVP_CipherFinal_ex(&ctx, outbuf, &outlen))
        {
		fprintf(stderr, "%s: Error finalizing the structure.\n", getprogname());
        	/* Cleans up the structure, erases key*/
                EVP_CIPHER_CTX_cleanup(&ctx);
                memset(key, 0, 16);
		exit(EXIT_FAILURE);
        }
        
	/*Writes the last part of the ciphertext.*/
	fwrite(outbuf, 1, outlen, out);

	/*Cleans up the structure to leave no trace.*/
        if(EVP_CIPHER_CTX_cleanup(&ctx) == 0)
        {
                fprintf(stderr, "%s: The clean up failed.\n", getprogname());
		/*Erases the key.*/
                memset(key,0,16);
                exit(EXIT_FAILURE);
        }
	
	/*Erases the key.*/
        memset(key,0,16);

	return EXIT_SUCCESS;
}

int main(int argc, char* argv[])
{
	setprogname("bfed");
	/*Sets the process title to bfed and skips executable name.*/
	setproctitle("-%s", "bfed");
	
	int encrypt_flag = -1;
	int key_flag = 0;
	char ch;
        unsigned char raw_key[16] = {'\0'};

	
	/*If there are no agruments passed, display help function.*/
	if (argc <=1)
		help();

	/*Get the options.*/
        while ( (ch = getopt(argc, argv,"dehk")) != -1 )
        {
                switch(ch)
                {
                        case 'd':
				encrypt_flag = 0;
				break;
			case 'e':
				encrypt_flag = 1;
				break;
			case 'h':
				help();
				break;
			case 'k':

				if (strlen(argv[optind]) > 16)
				{
					fprintf(stderr, "%s: The key is larger than 16 characters.\n", getprogname());
					exit(EXIT_FAILURE);
				}

				/*Get the key from the command line*/	
				memcpy(raw_key, argv[optind], 16);
				optind++;
				key_flag = 1;
				int index = 0;

				/*Checks to make sure the string is hex.*/
				for(index=0; index<16; index++)
					switch(raw_key[index])
					{
						case '0':
							break;
						case'1':
                                                        break;							
						case'2':
                                                        break;
						case'3':
                                                        break;
						case'4':
                                                        break;
						case'5':
                                                        break;
						case'6':
                                                        break;
						case'7':
                                                        break;
						case'8':
                                                        break;
						case'9':
                                                        break;
						case'a':                                                        
                                                        break;
						case'b':
                                                        break;
						case'c':
                                                        break;
						case'd':
                                                        break;
						case'e':                                                        
                                                        break;
						case'f':
                                                        break;
						case'A':
                                                        break;
                                                case'B':
                                                        break;
                                                case'C':
                                                        break;
                                                case'D':
                                                        break;
                                                case'E':
                                                        break;
                                                case'F':
                                                        break;
						default:
							fprintf(stderr, "%s: The key is not a hexidecimal key.\n", getprogname());
							exit(EXIT_FAILURE);
					}
				
				break;
			case '?':
				help();
				break;
			default:
				help();
				break;
		}//ends the switch statement for the options.
	}//ends the while loop parsing the options
 
	/*Checks to make sure that the key was entered and either -e or -d was specified.*/
	if((!key_flag) || (encrypt_flag == -1))
		help();

	/*calls the enc/dec algorithm with stdin and stdout as the file pointers.*/
	do_crypt(stdin,stdout,encrypt_flag, raw_key);

        return EXIT_SUCCESS;
}

